// SymbolTable.java
//
// ICS 142 Winter 2004
// Assignment #3
//
// A SymbolTable stores all of the declarations in a Monkie2004 program --
// variable declarations, types, and subprograms -- and provides fast lookup
// for each of them.  These are the methods that you will find to be of use:
//
//   public void enterScope()
//       Call this method whenever you enter a new scope during semantic
//       analysis.  This will allow variables to be redeclared in the new
//       scope.
//
//   public void exitScope()
//       Call this method whenever you exit a scope during semantic analysis.
//       This will cause any of the variables declared while analyzing the
//       scope to vanish.
//
//   public Type lookupType(String identifier)
//       Returns a Type object for the type of the given name, or null if
//       no such type exists.
//
//   public Subprogram lookupSubprogram(String identifier)
//       Returns a Subprogram object for the subprogram of the given name,
//       or null if no such subprogram exists.
//
//   public void declareSubprogram(String identifier, ArrayList parameterList,
//                                 boolean isFunction, Type returnType)
//       Declares a new subprogram, given a name, an ArrayList of Type objects
//       for the formal parameters (only the types of the formal parameters are
//       stored in the Subprogram object; the names are not), a boolean that
//       indicates whether it is a procedure or function, and a Type object
//       corresponding to the subprogram's return type, if any.  If the
//       subprogram is a procedure (and thus returns no type), returnType
//       should be null.
//
//   public Variable lookupVariable(String identifier)
//       Returns a Variable object for the variable of the given name, or
//       null if no such variable exists.  All scopes are searched, from the
//       innermost to the outermost.
//
//   public Variable lookupLocalVariable(String identifier)
//       Returns a Variable object for the variable of the given name if
//       declared in the current scope, or null if no such variable exists.
//       (Sometimes you need this method; for example, if you want to ensure
//       when analyzing a local variable declaration that a variable is not
//       already declared in the current scope.)
//
//   public void declareVariable(String identifier, Type type)
//       Declares a variable in the current scope, with the given identifier
//       and type.
//
// When created, the SymbolTable will consist of all of the predefined
// types and subprograms in Monkie2004.
//
// THIS CLASS SHOULD BE LEFT AS-IS!


import java.util.*;


public class SymbolTable
{
	private HashMap typeMap;
	private HashMap subprogramMap;
	private ArrayList variableMaps;


	public SymbolTable()
	{
		typeMap = new HashMap();

		declareType("integer");
		declareType("string");
		declareType("boolean");
		
		subprogramMap = new HashMap();

		declareSubprogram("read_string", new ArrayList(), true, lookupType("string"));

		ArrayList printStringParameterTypeList = new ArrayList();
		printStringParameterTypeList.add(lookupType("string"));
		declareSubprogram("print_string", printStringParameterTypeList, false, null);

		declareSubprogram("read_integer", new ArrayList(), true, lookupType("integer"));
		
		ArrayList printIntegerParameterTypeList = new ArrayList();
		printIntegerParameterTypeList.add(lookupType("integer"));
		declareSubprogram("print_integer", printIntegerParameterTypeList, false, null);

		declareSubprogram("read_boolean", new ArrayList(), true, lookupType("boolean"));
		
		ArrayList printBooleanParameterTypeList = new ArrayList();
		printBooleanParameterTypeList.add(lookupType("boolean"));
		declareSubprogram("print_boolean", printBooleanParameterTypeList, false, null);
		
		declareSubprogram("print_endline", new ArrayList(), false, null);

		variableMaps = new ArrayList();

		enterScope();
	}


	public void enterScope()
	{
		variableMaps.add(new HashMap());
	}
	
	
	public void exitScope()
	{
		variableMaps.remove(variableMaps.size() - 1);
	}


	public Type lookupType(String identifier)
	{
		return (Type) typeMap.get(identifier);
	}


	public void declareType(String identifier)
	{
		typeMap.put(identifier, new Type(identifier));
	}
	
	
	public Subprogram lookupSubprogram(String identifier)
	{
		return (Subprogram) subprogramMap.get(identifier);
	}


	public void declareSubprogram(
		String identifier, ArrayList parameterList,
		boolean isFunction, Type returnType)
	{
		Subprogram sub = new Subprogram(identifier, parameterList, isFunction, returnType);
		subprogramMap.put(identifier, sub);
	}
	
	
	public Variable lookupVariable(String identifier)
	{
		for (int i = variableMaps.size() - 1; i >= 0; --i)
		{
			HashMap scopeVariableMap = (HashMap) variableMaps.get(i);
			
			Variable v = (Variable) scopeVariableMap.get(identifier);
			
			if (v != null)
				return v;
		}
		
		return null;
	}


	public Variable lookupLocalVariable(String identifier)
	{
		HashMap localVariableMap = (HashMap) variableMaps.get(variableMaps.size() - 1);
		return (Variable) localVariableMap.get(identifier);
	}


	public void declareVariable(String identifier, Type type)
	{
		Variable v = new Variable(identifier, type);

		HashMap localVariableMap = (HashMap) variableMaps.get(variableMaps.size() - 1);
		localVariableMap.put(identifier, v);
	}
}
